########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Lib/Uri.py,v 1.49 2003/06/06 23:07:16 mbrown Exp $
"""
Classes and functions related to URI validation and resolution

Copyright 2003 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

import urlparse, urllib, os, re
from UserDict import UserDict
import Uuid
from Ft.Lib import UriException


# URI scheme for 4Suite repo resources
# (might someday be changed to something like
#  'vnd.ft.ftss' if proposed standards are adopted).
FTSS_URI_SCHEME = 'ftss'

# @@ FTSS_URI_SCHEME should be removed from here and added
# back by the appropriate driver
DEFAULT_URI_SCHEMES = ('http', 'ftp', 'file', 'gopher', FTSS_URI_SCHEME)

class UriDict(UserDict):
    """
    A dictionary that uses URIs as keys. It treats 'file:///' and
    'file://localhost/' as identical (RFC 1738 says this is a special
    case). Used by Ft.Xml.Xslt.XsltContext.
    """
    def __normalizekey(self, key):
        if key.startswith('file://localhost/'):
            return 'file://' + key[16:]
        else:
            return key

    def __getitem__(self, key):
        return self.data[self.__normalizekey(key)]

    def __setitem__(self, key, value):
        self.data[self.__normalizekey(key)] = value
        return

    def __delitem__(self, key):
        del self.data[self.__normalizekey(key)]

    def has_key(self, key):
        return self.data.has_key(self.__normalizekey(key))


#============================================================================
# Functions to inspect or validate URIs and URI references
#
def MatchesUriRefSyntax(s, strict=1):
    """
    This function returns 1 if a given string could be a URI reference (an
    absolute or relative URI, plus an optional #fragment) as defined in RFC
    2396, just based on the string's syntax.

    The 'strict' argument can be set to 0 for lenient, faster checking.

    Note that an empty string is a valid URI reference. Also note that a
    successful match does not guarantee that the string is a URI reference;
    only that it could be one, since it meets the syntax requirements.
    Failure to match indicates that it definitely is not a URI reference.
    """
    _initUriValidationRegex()
    if strict:
        return pyStrictUriRefPattern.match(s) is not None
    else:
        return pyUriRefPattern.match(s) is not None

def MatchesUriSyntax(s, strict=1):
    """
    This function returns 1 if a given string could be an absolute or
    relative URI, as defined in RFC 2396, just based on the string's syntax.

    The 'strict' argument can be set to 0 for lenient, faster checking.

    Note that an empty string is a valid URI. Also note that a successful
    match does not guarantee that the string is a URI; only that it could
    be one, since it meets the syntax requirements. Failure to match
    indicates that it is definitely not a URI.
    """
    _initUriValidationRegex()

    if strict:
        return pyStrictUriPattern.match(s) is not None
    else:
        return pyUriPattern.match(s) is not None

def IsAbsolute(identifier):
    """
    Given a string believed to be a URI or URI reference, tests that the
    identifier is absolute (as per RFC 2396), not relative -- i.e., that
    it has a scheme.
    """
    # We do it this way to avoid compiling another massive regex.
    return SplitUriRef(identifier)[0] is not None

def SplitUriRef(uriref):
    """
    Given a valid URI reference as a string, split it into a tuple
    with the following structure, as per RFC 2396 appendix B, with
    the addition of opaque_part to avoid confusion over the meaning
    of 'path':

    (scheme, authority, path, query, fragment, opaque_part)

    All values will be strings (possibly empty) or None. If there
    is an opaque_part, then there will be no authority, path, or
    query.
    """

    _initSplitUriRefPattern()
    # get scheme, authority, path, query, fragment
    g = splitUriRefPattern.match(uriref).groupdict()
    scheme      = g['scheme']
    authority   = g['authority']
    path        = g['path']
    query       = g['query']
    fragment    = g['fragment']

    # if there is a scheme and the path isn't absolute, then
    # there is an opaque part, not an authority + path + query
    opaque_part = None
    if g['scheme'] and not g['path'].startswith('/'):
        opaque_part = uriref[len(scheme)+1:]
        if fragment:
            opaque_part = opaque_part[:-len(fragment)-1]
        authority = None
        path = None
        query = None

    return (scheme, authority, path, query, fragment, opaque_part)


#============================================================================
# Functions to resolve relative URI references to absolute form
#
def Absolutize(uriRef, docUri, baseUri=None):
    """
    Resolves a URI reference to absolute form, effecting the result of
    RFC 2396 section 5.2. Functionally equivalent to urlparse.urljoin()
    and urllib.basejoin(), but without the bugs. It is the caller's
    responsibility to only call this when the base URI's scheme is one
    that uses a hierarchical path component (e.g., http, ftp, and
    usually file, although the file scheme is generally treated with
    a great degree of leniency).

    Unless provided, the base URI is assumed to be the document URI.
    An exception is raised if the base URI or document URI are not
    absolute.

    If it is determined that absolute form of the reference, minus the
    fragment identifier, is the document URI, then None is returned,
    indicating to the caller that the reference can be satisfied within
    the context of the local, at-hand representation of the resource;
    the base URI does not apply, and a new representation of the
    resource should not be sought. (Returning a URI would be more
    accurate, but returning None facilitates a boolean test, whereas
    returning a URI would place unnecessary burden on the caller).

    Example:
      # current_doc = object wrapping a resource representation & URI
      # href = URI reference found in the document
      # base = base URI defined in the document or transport protocol
      doc = current_doc
      frag = SplitUriRef(href)[4]
      fullUri = Absolutize(href, doc.uri, base)
      if fullUri:
        # it's some other doc; go use it
        # (assumes the resolver will discard the frag, if any)
        docStream = BASIC_RESOLVER.resolve(fullUri)
        # get a new object (foo is XML parser or whatever)
        doc = foo(docStream, fullUri)
      if frag:
        # use the part of the doc identified by the frag
      else:
        # use the whole doc
    """
    # The bugs referred to above include:
    # 1. urllib.basejoin() mishandles the '' reference.
    # 2. Both omit an empty authority component from 'file' URIs.
    # 3. Both fail to distinguish between the base and the document.
    if baseUri is None:
        baseUri = docUri
    if not docUri or not IsAbsolute(docUri):
        raise UriException(UriException.RELATIVE_DOCUMENT_URI,
                           repr(docUri))
    if not baseUri or not IsAbsolute(baseUri):
        raise UriException(UriException.RELATIVE_BASE_URI,
                           repr(baseUri))
    # (2) shortcut for the simplest "current doc" cases
    if uriRef == '' or uriRef[0] =='#':
        return None
    # (3) absolute URI references are not relative to anything.
    # no workaround for refs with scheme same as base.
    (rScheme, rAuth, rPath, rQuery, rFrag, rOpq) = SplitUriRef(uriRef)
    if rScheme is not None:
        return uriRef
    # if the base has an opaque part, then the mechanism for resolving
    # to absolute form is undefined.
    (bScheme, bAuth, bPath, bQuery, bFrag, bOpq) = SplitUriRef(baseUri)
    if bOpq:
        raise UriException(UriException.OPAQUE_BASE_URI,
                           repr(baseUri), repr(uriRef))
    # ok, base is hierarchical... start assembling the new URI
    uri = ''
    # if ref path is an absolute-path or network-path, we'll be using it
    nPath = rPath
    # special case: file://localhost/ == file:/// per RFC 1738
    if bScheme == 'file' and bAuth == 'localhost':
        bAuth = ''
    # (2) base URI query and fragment are ignored
    del bQuery, bFrag
    # (3) if no scheme, inherit scheme from base URI
    if not rScheme:
        rScheme = bScheme
    # (4) if authority undefined, inherit from base URI, then examine path
    if not rAuth:
        rAuth = bAuth
        # (5) if ref path doesn't start with '/', ref is a relative-path
        # rPath could be empty string if query is present
        if (rQuery and not rPath) or (rPath and rPath[0] != '/'):
            # (6a) copy all but last segment of base URI's path
            # (6b) append ref path
            # account for possibility of bPath being None
            if bPath:
                nPath = bPath[:bPath.rfind('/')+1] + rPath
            else:
                nPath = rPath
            # (6c) remove './' from all '/./'
            nPath = nPath.replace('/./', '/')
            # (6d) remove '.' from '/.' at end
            if nPath[-2:] == '/.':
                nPath = nPath[:-1]
            # (6e,6f,6g,6h) remove segments followed by '..',
            # and remove extraneous '..' components
            nPath = _eliminateBackRefs(nPath)
    # (4) authority component is defined, so ref is a network-path; or
    # (5) ref path starts with '/', so ref is an absolute-path.
    # (7) assemble new URI
    if rScheme is not None:
        uri = rScheme + ':'
    if rAuth is not None:
        uri += '//' + rAuth
    uri += nPath
    if rQuery is not None:
        uri += '?' + rQuery
    # resolved URI ended up being current doc?
    if uri == docUri:
        return None
    if rFrag is not None:
        uri += '#' + rFrag
    return uri

#============================================================================
# Private functions and variables that support the functions above
#
_backRefSetupCompleted = 0
def _initBackRefRegex():
    """
    Called internally to compile the regular expression needed by
    Absolutize(), just once, the first time that function is called.
    """
    global _backRefSetupCompleted
    if _backRefSetupCompleted:
        return
    escaped         = r"(?:%[0-9a-fA-F]{2})"
    rel_segment     = r"(?:[a-zA-Z0-9\-_.!~*'();@&=+$,]|%s)+" % escaped
    global _backRefPattern
    _backRefPattern = re.compile(r'/' + rel_segment + r'/\.\.(/.*)')
    _backRefSetupCompleted = 1

def _eliminateBackRefs(absPath):
    """
    Supports Absolutize(). Given an absolute path, recursively removes
    '..' backreferences. Extraneous backreferences that would go 'above'
    the root are removed.
    """
    _initBackRefRegex()
    newPath = _backRefPattern.sub(r'\1', absPath, 1)
    if newPath != absPath:
        return _eliminateBackRefs(newPath)
    else:
        if newPath[-3:] == '/..':
            newPath = newPath[:newPath[:-3].rfind('/')+1] or '/'
        return newPath.replace('/../','/')

_validationSetupCompleted = 0
def _initUriValidationRegex():
    """
    Called internally to compile the regular expressions needed by
    URI validation functions, just once, the first time a function
    that needs them is called.
    """
    global _validationSetupCompleted
    if _validationSetupCompleted:
        return
    _validationSetupCompleted = 1

    #========================================================================
    # Regular expressions for determining the non-URI-ness of strings
    #
    # 1. Altova's regex
    #
    # We don't use this one; it's just here for reference.
    #
    # Courtesy Altova (in the public domain)
    # hacked from the BNF in RFC 2396 (http://www.ietf.org/rfc/rfc2396.txt)
    # URI_REGEX = r"(([a-zA-Z][0-9a-zA-Z+\-\.]*:)?/{0,2}[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?(#[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?"
    #
    # Altova's regex is actually very lenient and will match either a URI or
    # a URI reference anywhere in a string. Since an empty string is a valid
    # URI (it is zero or more URI-looking things), this results in *every*
    # string matching the pattern! So, in order for a pattern that possibly
    # matches an empty string to be useful, we have to be sure the pattern
    # matches the entire string we are testing. So don't look for a URI ref
    # *in* a string, look for a string that is, from start to finish, a URI
    # ref, like this: \A(?!\n)URI_REGEX\Z. The \A\Z delimiters take care of
    # matching the whole string, and the (?!\n) ensures that the edge case
    # of a sole linefeed character won't cause a match.
    #
    # Also, one must be careful to realize that a test like this only
    # determines whether something matches the URI reference syntax, and
    # is not a guarantee that it is actually a URI; it could just be an
    # alphabetic string like "helloWorld" or "C:\autoexec.bat". All you can
    # say with this test is that it definitely is not a URI, or that it
    # "could be" a URI.
    #
    # As it might be more useful to test a string to see if it could be just
    # a URI and not a URI reference, we'll derive from the Altova regex a
    # pair of Python regular-expression objects, one for URIs and the other
    # for URI references. We optimize it slightly by using "(?:foo)" instead
    # of "(foo)" for grouping.
    #
    #
    # 2. Python regular expressions for strict validation of URI references
    #    by Mike Brown (Fourthought, Inc.)
    #
    # Public domain.
    #
    # Based on the BNF in RFC 2396, as amended by RFC 2732
    #
    # version: 2002-02-18T21:18:00
    #
    # The compiled patterns use the Python-specific feature of named groups,
    # like "(?P<foo>bar)", so that the portion of the string matching the
    # 'foo' group's pattern can be referenced later. This means one can use a
    # Match object to see how the string was interpreted, and to access some
    # of its components, if desired, although it may be easier to just reparse
    # with SplitUriRef(). These patterns also use discardable groups,
    # indicated by "(?:foo)", for efficiency.
    #
    # Note that we do not use any \d or \w shortcuts, as these are
    # potentially locale or Unicode sensitive.
    #
    # A non-Python version of the expressions could be made by doing a simple
    # replacement, in the strings below
    #    "(?:"  becomes  "("
    #    "(?P<foo>"  where foo matches [a-zA-Z_]+  becomes  "("
    # However, there are about 170 groups. This might overflow a regular
    # expression parsing engine. Python only accepts 100.
    #

    URI_REGEX = r"(?:(?:[a-zA-Z][0-9a-zA-Z+\-\.]*:)?/{0,2}[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?"
    FRAGMENT_REGEX = r"(?:#[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?"
    URIREF_REGEX = "\A(?!\n)%s%s\Z" % (URI_REGEX, FRAGMENT_REGEX)
    URI_REGEX = "\A%s\Z" % (URI_REGEX)

    global uriPattern, uriRefPattern
    uriPattern = re.compile(URI_REGEX)           # lenient checking for URIs
    uriRefPattern = re.compile(URIREF_REGEX)     # lenient checking for URI refs

    #Build all of the pieces of the pattern
    escaped         = r"(?:%[0-9a-fA-F]{2})"
    uric            = r"(?:[;/?:@&=+$,[\]a-zA-Z0-9\-_.!~*'()]|%s)" % escaped # "[" and "]" per RFC 2732
    fragment        = r"(?P<fragment>%s*)" % uric
    query_A         = r"(?P<query_A>%s*)" % uric
    query_R         = r"(?P<query_R>%s*)" % uric
    pchar           = r"(?:[a-zA-Z0-9\-_.!~*'():@&=+$,]|%s)" % escaped
    param           = r"%s*" % pchar
    segment         = r"%s*(?:;%s)*" % (pchar, param)
    path_segments   = r"%s(?:/%s)*" % (segment, segment)
    port            = r"(?:[0-9]*)"
    IPv4address     = r"(?:[0-9]{1,3}\.){3}[0-9]{1,3}"
    hexseq          = r"(?:[0-9a-fA-F]{1,4}(?::[0-9a-fA-F]{1,4})*)"
    hexpart         = r"(?:%s|%s::%s?|::%s?)" % (hexseq, hexseq, hexseq, hexseq)
    IPv6address     = r"%s(?::%s)?" % (hexpart, IPv4address)
    ipv6reference   = r"\[%s\]" % IPv6address
    toplabel        = r"[a-zA-Z](?:[a-zA-Z0-9\-]*[a-zA-Z0-9])?"
    domainlabel     = r"[a-zA-Z0-9](?:[a-zA-Z0-9\-]*[a-zA-Z0-9])?"
    hostname        = r"(?:%s\.)*%s\.?" % (domainlabel, toplabel)
    host            = r"(?:%s|%s|%s)" % (hostname, IPv4address, ipv6reference)
    hostport_A      = r"(?P<hostport_A>%s(?::%s)?)" % (host, port)
    hostport_R      = r"(?P<hostport_R>%s(?::%s)?)" % (host, port)
    userinfo_A      = r"(?P<userinfo_A>(?:[a-zA-Z\-_.!~*'();:&=+$,]|%s)*)" % escaped
    userinfo_R      = r"(?P<userinfo_R>(?:[a-zA-Z\-_.!~*'();:&=+$,]|%s)*)" % escaped
    server_A        = r"(?:(?:%s@)?%s)" % (userinfo_A, hostport_A)
    server_R        = r"(?:(?:%s@)?%s)" % (userinfo_R, hostport_R)
    reg_name        = r"(?:[a-zA-Z0-9\-_.!~*'()$,;:@&=+]|%s)+" % escaped
    authority_A     = r"(?P<authority_A>%s|%s)" % (server_A, reg_name)
    authority_R     = r"(?P<authority_R>%s|%s)" % (server_R, reg_name)
    scheme          = r"(?P<scheme>[a-zA-Z][a-zA-Z0-9+\-.]*)"
    rel_segment     = r"(?:[a-zA-Z0-9\-_.!~*'();@&=+$,]|%s)+" % escaped
    abs_path_noId   = r"(?:/%s)" % path_segments
    abs_path_A      = r"(?P<abs_path_A>/%s)" % path_segments
    abs_path_R      = r"(?P<abs_path_R>/%s)" % path_segments
    rel_path        = r"(?P<rel_path>%s%s?)" % (rel_segment, abs_path_noId)
    net_path_A      = r"(?P<net_path_A>//%s%s?)" % (authority_A, abs_path_noId)
    net_path_R      = r"(?P<net_path_R>//%s%s?)" % (authority_R, abs_path_noId)
    uric_no_slash   = r"(?:[a-zA-Z0-9\-_.!~*'();?:@&=+$,]|%s)" % escaped # no "[" or "]": RFC 2732 oversight?
    opaque_part     = r"(?P<opaque_part>%s%s*)" % (uric_no_slash, uric)
    hier_part       = r"(?P<hier_part>(?:%s|%s)(?:\?%s)?)" % (net_path_A, abs_path_A, query_A)
    relativeURI     = r"(?P<relativeURI>(?:%s|%s|%s)(?:\?%s)?)" % (net_path_R, abs_path_R, rel_path, query_R)
    absoluteURI     = r"(?P<absoluteURI>%s:(?:%s|%s))" % (scheme, hier_part, opaque_part)
    URI             = r"(?P<URI>%s|%s)" % (absoluteURI, relativeURI)
    URI_reference   = r"(?P<URI_reference>%s?(?:#%s)?)" % (URI, fragment)
    STRICT_URI_PYREGEX = r"\A%s\Z" % URI
    STRICT_URIREF_PYREGEX = r"\A(?!\n)%s\Z" % URI_reference

    global pyStrictUriPattern, pyStrictUriRefPattern
    pyStrictUriPattern = re.compile(STRICT_URI_PYREGEX)       # strict checking for URIs
    pyStrictUriRefPattern = re.compile(STRICT_URIREF_PYREGEX) # strict checking for URI refs


_splitSetupCompleted = 0
def _initSplitUriRefPattern():
    """
    Called internally to compile the regular expression used by
    SplitUriRef() just once, the first time the function is called.
    """
    global _splitSetupCompleted
    if _splitSetupCompleted:
        return
    _splitSetupCompleted = 1

    # Like the others, this regex is also in the public domain.
    SPLIT_URIREF_REGEX = r"^(?:(?P<scheme>[^:/?#]+):)?(?://(?P<authority>[^/?#]*))?(?P<path>[^?#]*)(?:\?(?P<query>[^#]*))?(?:#(?P<fragment>.*))?$"
    global splitUriRefPattern
    splitUriRefPattern = re.compile(SPLIT_URIREF_REGEX)


uriSchemePattern = re.compile(r'\A([a-zA-Z][a-zA-Z0-9+\-.]*):')
def _getScheme(uriRef):
    """
    Called internally to get just the scheme from a URI reference, with
    the greatest efficiency.
    """
    # Using a regex seems to be the best option. Called 50,000 times on
    # different URIs, on a 1.0-GHz PIII with FreeBSD 4.7 and Python
    # 2.2.1, this method completed in 0.95s, and 0.05s if there was no
    # scheme to find. By comparison,
    #   urllib.splittype()[0] took 1.5s always;
    #   Ft.Lib.Uri.SplitUriRef()[0] took 2.5s always;
    #   urlparse.urlparse()[0] took 3.5s always.
    matches = uriSchemePattern.search(uriRef)
    if matches is None:
        return None
    else:
        return matches.groups()[0]


#============================================================================
# Extendable normalization and resolution functions for URI references
#
class BaseUriResolver:
    """
    This is the base class that provides a set of functions related to the
    resolution of URIs.
    """
    def __init__(self):
        self.supportedSchemes = list(DEFAULT_URI_SCHEMES)

    def normalize(self, uri, base):
        """
        Given a URI reference that is either absolute or relative, and
        an absolute base URI, this function produces an absolute URI
        reference, effecting the result of the algorithm described in
        section 5.2 of RFC 2396 ("Resolving Relative References to Absolute
        Form"), and also verifying that the resulting URI reference has a
        scheme that resolve() supports.

        According to the RFC, the base URI is required to have a scheme,
        but for convenience, this function allows for the possibility of
        the base URI beginning with a '/', in which case the 'file:' scheme
        is assumed. A UriException is raised if the normalized URI
        references an unsupported scheme or is otherwise invalid.
        """
        # check that base is valid
        if base is not None and base != '':
            # assume file: if leading "/"
            if base[0] == '/':
                base = 'file://' + base
            base_scheme = _getScheme(base)
            # although it's possible to normalize other schemes like https,
            # the default should be just the ones resolve() handles
            if base_scheme in self.supportedSchemes:
                uri = Absolutize(uri, base)
                if not uri:
                    return base
            # base has unrecognized scheme
            else:
                raise UriException(UriException.INVALID_BASE_URI, base)
            # base is OK
            return uri
        # base is not valid
        else:
            raise UriException(UriException.INVALID_BASE_URI, base)

    def resolve(self, uri, base=None):
        """
        This function takes an absolute or relative URI, and an optional
        base URI, produces a normalized URI using the normalize function,
        attempts to obtain access to an entity representing the resource
        identified by the URI, and returns the entity as a stream. If the
        stream could not be obtained, an exception is raised. The default
        URI schemes supported are http, ftp, gopher, file and %s.
        """ % FTSS_URI_SCHEME
        if base:
            uri = self.normalize(uri, base)
        (scheme, authority, path, query, fragment, opaque_part) = SplitUriRef(uri)
        if scheme == 'file':
            path = UriToOsPath(uri)
            try:
                stream = open(path, 'rb')
            except IOError, e:
                raise UriException(UriException.RESOURCE_ERROR, path, str(e))
        else:
            try:
                stream = urllib.urlopen(uri)
            except IOError, e:
                raise UriException(UriException.RESOURCE_ERROR, uri, str(e))
        return stream

    def splitFragment(self, uri):
        frag_split = uri.split("#")
        if len(frag_split) == 2:
            return frag_split[0], frag_split[1]
        else:
            return frag_split[0], ""

    def generate(self, hint=None):
        """
        This function generates and returns a URI.
        The hint is an object that helps decide what to generate.
        The default action is to generate a random UUID URN.
        """
        return 'urn:uuid:'+Uuid.UuidAsString(Uuid.GenerateUuid())


class SchemeRegistryResolver(BaseUriResolver):
    """
    A type of resolver that allows developers to register different callable
    objects to handle different URI schemes.  The default action if there
    is nothing registered for the scheme will be to fall back to
    BaseUriResolver behavior *unless* you have in the mapping a special
    scheme None.  The callable object that is the value on that key will
    then be used as the default for all unknown schemes.

    The expected function signature for scheme call-backs
    matches BaseUriResolver.resolve:

    resolve(self, uri, base=None)

    You can manipulate the mapping directly using the "handlers" attribute
    """
    def __init__(self, handlers=None):
        """
        handlers - a Python dictionary with scheme names as keys (e.g. "http")
        and callable objects as values
        """
        BaseUriResolver.__init__(self)
        self.handlers = handlers or {}
        return

    def resolve(self, uri, base=None):
        scheme = _getScheme(uri)
        if not scheme:
            if base:
                scheme = _getScheme(base)
            if not scheme:
                #Another option is to fall back to Base class behavior
                raise Exception('SchemeRegistryResolver can only handle URIs with a scheme')
        func = self.handlers.get(scheme)
        if not func:
            func = self.handlers.get(None)
            if not func:
                return BaseUriResolver.resolve(self, uri, base)
        return func(uri, base)


BASIC_RESOLVER = BaseUriResolver()


def OsPathToUri(path, attemptAbsolute=0):
    """
    This function takes an OS path and forces it to a URL of the form
    file:///path/to/file.ext

    Warning: Such a file URL might not behave as desired when
    passed to some libraries.

    If the given path is relative, it will result in a URL of the
    dubious form file:relative/path/to/file.ext, unless the
    attemptAbsolute flag is set, in which case the current working
    directory will be used as the basis for attempting to make the
    URL take the first form.
    """
    # On Windows, urllib.pathname2url() doesn't handle NetBIOS names
    # very well. It also prepends drivespecs with ///. It also removes
    # trailing slashes. We try to work around those things here.
    host = ''
    if os.name == 'nt' and path.startswith('\\\\'):
        host = path[2:path[2:].find('\\')+2]
        path = path[len(host)+2:]
    path = urllib.pathname2url(path) + '/' * (os.name == 'nt' and
                                              path.endswith('\\'))
    if path.startswith('///'):
        return 'file:' + path
    elif path.startswith('/'):
        return 'file://' + host + path
    else:
        if attemptAbsolute:
            current_path = urllib.pathname2url(os.getcwd())
            if current_path.startswith('///'):
                current_path = current_path[2:]
            return 'file://' + current_path + '/' + path
        else:
            return 'file:' + path


def UriToOsPath(uri, attemptAbsolute=0):
    """
    This function takes a file URI (absolute or relative) and
    formats it as an OS path appropriate for the local system.
    If the given path is relative, the attemptAbsolute flag will
    determine whether the returned path is relative or absolute
    (absolute being based on the current working directory). As
    the name indicates, this is is just an attempt, and may give
    the wrong result.
    """
    #print "URI: %r" % uri
    scheme = _getScheme(uri)
    if scheme:
        if scheme == 'file':
            uri = uri[5:]
        else:
            raise UriException(UriException.NON_FILE_URI, uri)
    # chop off fragment, if any
    uri = re.sub(r'(.*)#.*', r'\1', uri)
    # file://localhost/foo is a special case. It is *always*
    # on the local system, per RFC 1738 and RFC 1630.
    if uri.startswith('//localhost/'):
        uri = uri[11:]
    path = urllib.url2pathname(uri)
    # Is there anything in urllib that works right? On Windows,
    # url2pathname('///path/to/file') becomes r'\\\path\to\file',
    # which is not treated the same as r'\path\to\file' by things
    # like open() or os.access()... so we clean that up here.
    if os.name == 'nt' and path.startswith('\\\\\\'):
        path = path[2:]
    # url2pathname('//<host>/path/to/file') on Unix comes back
    # unchanged! If host is not empty or localhost, it ought to raise
    # an exception, because there's no access method defined for remote
    # hosts on Unix. On Windows we can assume NetBIOS, which is why we
    # leave the host in the path on that platform.
    # No idea about other platforms, though.
    if os.name == 'posix' and path.startswith('//'):
        # ///path/to/file (empty host) is OK
        if path[2] == '/':
            path = path[2:]
        else:
            raise UriException(UriException.UNIX_REMOTE_HOST_FILE_URI, uri)
    if attemptAbsolute and not os.path.isabs(path):
        path = os.path.join(os.getcwd(), path)
    #print "OS path: %r" % path
    return path


def PathResolve(paths):
    """
    This function takes a list of file URIs.  The first can be
    absolute or relative to the URI equivalent of the current working
    directory. The rest must be relative to the first.
    The function converts them all to OS paths appropriate for the local
    system, and then creates a single final path by resolving each path
    in the list against the following one. This final path is returned
    as a URI.
    """
    if not paths: return paths
    paths = [UriToOsPath(p) for p in paths]
    if not os.path.isabs(paths[0]):
        paths[0] = os.path.join(os.getcwd(), paths[0])
    resolved = reduce(lambda a, b: \
                       BaseJoin(os.path.isdir(a)
                                 and OsPathToUri(
                                      os.path.join(a, '')
                                     ) or OsPathToUri(a),
                                OsPathToUri(b)[5:]),
                       paths)
    return resolved


def BaseJoin(base, rel):
    """
    This function is really just an abstraction of urllib.basejoin for now,
    because of our leeriness of the urllib and urlparse modules in Python
    """
    import urllib
    uri = urllib.basejoin(base, rel)
    if uri.startswith('file:/') and uri[6] != '/':
        # just a path portion, add empty host
        uri = 'file://' + uri[5:]
    return uri

#============================================================================
#
# see the following XML-SIG threads for 'file:' URI resolution discussion
#> http://aspn.activestate.com/ASPN/Mail/Message/xml-sig/614905
#> http://aspn.activestate.com/ASPN/Mail/Message/xml-sig/614913
#> http://mail.python.org/pipermail/xml-sig/2001-February/004572.html
#
# also see how Perl handles them:
#  http://www.perldoc.com/perl5.8.0/lib/URI/file.html
#
#============================================================================
#
# These are valid absoluteURIs:
#
#  file:/autoexec.bat     (scheme ":" abs_path)
#  file:///autoexec.bat   (scheme ":" net_path)
#
# This one is valid but is not what you'd expect it to be:
#
#  file://autoexec.bat    (authority = autoexec.bat, no abs_path)
#
# If you have any more than 3 slashes, it's OK because each path segment
# can be an empty string.
#
# This one is valid too, although everything after 'file:' is
# considered an opaque_part:
#
#   file:etc/passwd
#
# Unescaped backslashes are NOT allowed in URIs, ever.
# It is not possible to use them as path segment separators.
# Yet... Windows Explorer will accept these:
#   file:C:\WINNT\setuplog.txt
#   file:/C:\WINNT\setuplog.txt
#   file:///C:\WINNT\setuplog.txt
# However, it will also accept "|" in place of the colon after the drive:
#   file:C|/WINNT/setuplog.txt
#   file:/C|/WINNT/setuplog.txt
#   file:///C|/WINNT/setuplog.txt
#
# RFC 1738 says file://localhost/ and file:/// are equivalent;
# localhost in this case is always the local machine, no matter what
# your DNS says.
#
# Basically, any file: URI is valid. Good luck resolving them, though.
#
# Jeremy's idea is to not use open() or urllib.urlopen() on Windows;
# instead, use a C function that wraps Windows' generic open function,
# which resolves any path or URI exactly as Explorer would (he thinks)


#============================================================================
# Unused utility for cleaning up matches from strict URI validation
#
#def _normalizeStrictPatternGroups(matchGroupDict):
#    """
#    Given a dictionary as obtained with m.groupdict(), where m is the Match
#    object returned by pyStrictUriRefPattern.match(s) or
#    pyStrictUriPattern.match(s), returns a dictionary in which all keys
#    with None values have been removed, and in which the _A or _R suffixes
#    on the keys have been removed.
#
#    (These suffixes exist because of the need to keep group names in the
#    regular expressions unique, while still being able to use them multiple
#    times within the expression).
#
#    Example usage:
#
#    from Ft.Lib import Uri
#    s = 'http://u:pass@host.foo:8080/stuff/doc.html?up=down&left=right#toc'
#    m = Uri.pyStrictUriRefPattern.match(s)
#    dict = Uri._normalizeStrictPatternGroups(m.groupdict())
#    groups = dict.keys()
#    groups.sort()
#    for group in groups:
#        print "%s: %s" % (group, dict[group])
#
#    This will give you:
#
#    URI: http://u:pass@host.foo:8080/stuff/doc.html?up=down&left=right
#    URI_reference: http://u:pass@host.foo:8080/stuff/doc.html?up=down&left=right#toc
#    absoluteURI: http://u:pass@host.foo:8080/stuff/doc.html?up=down&left=right
#    authority: u:pass@host.foo:8080
#    fragment: toc
#    hier_part: //u:pass@host.foo:8080/stuff/doc.html?up=down&left=right
#    hostport: host.foo:8080
#    net_path: //u:pass@host.foo:8080/stuff/doc.html
#    port: 8080
#    query: up=down&left=right
#    scheme: http
#    userinfo: u:pass
#    """
#    r = {}
#    for group in matchGroupDict.keys():
#        if matchGroupDict[group] is not None:
#            newgroup = group
#            if group.endswith('_A') or group.endswith('_R'):
#                newgroup = group[:-2]
#            r[newgroup] = matchGroupDict[group]
#    return r
