########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Xml/Domlette.py,v 1.32 2003/05/01 05:48:18 uogbuji Exp $
"""
Abstraction module for the Domlettes.  Domlettes are DOM-like libraries
tailored for use in XPath/XSLT
cDomlette is used as much as possible

Copyright 2003 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

import os, sys, cStringIO, warnings
import InputSource
import codecs, encodings
from types import UnicodeType

from Ft.Xml import READ_EXTERNAL_DTD


# Common routines regardless of environment
from Lib.Nss import GetAllNs, SeekNss
from Lib.Print import Print, PrettyPrint

try:
    from Lib.XmlStripc import *
except ImportError:
    from Lib.XmlStrip import *

MINIDOM = int(os.environ.get("USE_MINIDOM", "0"))
if MINIDOM:
    from FtMiniDom import implementation, nonvalParse, valParse

else:
    from cDomlette import implementation, nonvalParse, valParse


def ConvertDocument(oldDocument, documentURI=u''):
    """
    Since foreign DOMs are not supported within 4Suite, this function
    lets users create a Domlette DOM from whatever DOM they are using.

    If the documentURI is not specified, it will try and get it from
    the document using DOM L3 attributes documentURI, then baseURI. If
    no URI is found, a warning is issued and an urn:uuid is generated
    for it.
    """
    if not documentURI:
        # DOM Level 3 URI support only
        if hasattr(oldDocument, 'documentURI'):
            documentURI = oldDocument.documentURI
        elif hasattr(oldDocument, 'baseURI'):
            documentURI = oldDocument.baseURI
        else:
            from Ft.Lib import Uuid
            documentURI = 'urn:uuid:'+Uuid.UuidAsString(Uuid.GenerateUuid())
            warnings.warn("Document created without an URI",
                          RuntimeWarning, 2)

    document = implementation.createRootNode(documentURI)
    for oldChild in oldDocument.childNodes:
        child = document.importNode(oldChild, 1)
        document.appendChild(child)

    return document


class _Reader:
    """
    Now there is one reader to do everything
    """
    def __init__(self, parseMethod, inputSourceFactory=None, args=(), kwargs=None):
        self.inputSourceFactory = inputSourceFactory or InputSource.DefaultFactory
        self.parseMethod = parseMethod
        self.kwargs = kwargs or {}
        self.args = args
        return

    def parse(self, inputSource):
        """
        Parse in a input source and return a DOM
        """
        if not isinstance(inputSource, InputSource.InputSource):
            raise TypeError("inputsource must be an instance of "
                            "Ft.Xml.InputSource.InputSource")
        return apply(self.parseMethod, (inputSource,) + self.args, self.kwargs)

    def parseString(self, st, uri=None, *v_args, **kw_args):
        isrc = apply(self.inputSourceFactory.fromString, (st, uri) + v_args, kw_args)
        return self.parse(isrc)

    def parseStream(self, stream, uri=None, *v_args, **kw_args):
        isrc = apply(self.inputSourceFactory.fromStream, (stream, uri) + v_args, kw_args)
        return self.parse(isrc)

    def parseUri(self, uri, *v_args, **kw_args):
        isrc = apply(self.inputSourceFactory.fromUri, (uri,) + v_args, kw_args)
        return self.parse(isrc)


class ValidatingReaderBase(_Reader):
    """
    Base class to be used by all validating readers
    """
    def __init__(self, inputSourceFactory=None):
        _Reader.__init__(self, valParse, inputSourceFactory)


class NonvalidatingReaderBase(_Reader):
    """
    Base class to be used by all non validating readers
    """
    def __init__(self, inputSourceFactory=None):
        _Reader.__init__(self, nonvalParse, inputSourceFactory,
                         kwargs={'readExtDtd': READ_EXTERNAL_DTD})


class NoExtDtdReaderBase(_Reader):
    """
    Base class to be used by all non validating readers that do not even read
    the external subset
    """
    def __init__(self, inputSourceFactory=None):
        _Reader.__init__(self, nonvalParse, inputSourceFactory,
                         kwargs={'readExtDtd': 0})


#Create the instances of them
ValidatingReader = ValidatingReaderBase()
NonvalidatingReader = NonvalidatingReaderBase()
NoExtDtdReader = NoExtDtdReaderBase()

try:
    from Ft.Xml import Catalog
    CatalogValidatingReader = ValidatingReaderBase(
        inputSourceFactory=Catalog.DefaultCatalogInputSourceFactory)
    CatalogNonvalidatingReader = NonvalidatingReaderBase(
        inputSourceFactory=Catalog.DefaultCatalogInputSourceFactory)
except Exception, e:
    class ErrorReader(_Reader):
        def parse(self, inputSource):
            raise RuntimeError("You must have PyExpat installed to use Catalogs")

    CatalogValidatingReader = ErrorReader(None)
    CatalogNonvalidatingReader = ErrorReader(None)


class DeprecatedReader:
    """
    Defined to support the old interfaces
    """
    def __init__(self,resolveEntity = None, processIncludes = 1):
        warnings.warn("You are using deprecated readers",
                      DeprecationWarning, 2)
        self.resolveEntity = resolveEntity
        self.processIncludes = processIncludes
        self.inputSourceFactory = InputSource.DefaultFactory

    def fromUri(self, uri, baseUri='', ownerDoc=None, stripElements=None):
        """
        Create a default input Source
        """
        if not baseUri:
            warnings.warn("Creation of InputSource without an URI",
                          RuntimeWarning, 2)
        src = InputSource.InputSource(None, uri,
                                      processIncludes=self.processIncludes,
                                      stripElements=stripElements)
        src = src.resolve(uri, '', '')
        return self._parseMethod[0](src)

    def fromStream(self, stream, refUri='', ownerDoc=None,
                   stripElements=None):
        """
        Create an input source
        """
        if not refUri:
            warnings.warn("Creation of InputSource without an URI",
                          RuntimeWarning, 2)
        src = InputSource.InputSource(stream, refUri,
                                      processIncludes=self.processIncludes,
                                      stripElements=stripElements)
        return self._parseMethod[0](src)


    encoder = codecs.lookup("utf-8")[0] # encode,decode,reader,writer
    def fromString(self, st, refUri='', ownerDoc=None, stripElements=None):
        "Create a DOM from a string"

        if isinstance(st, UnicodeType):
            st = self.encoder(st)[0]
        stream =  cStringIO.StringIO(st)

        if not refUri:
            warnings.warn("Creation of InputSource without an URI",
                          RuntimeWarning, 2)
        rt = self.fromStream(stream, refUri, ownerDoc, stripElements)
        stream.close()
        return rt

class DEFAULT_NONVALIDATING_READER(DeprecatedReader):
    _parseMethod = (nonvalParse,)

class DEFAULT_VALIDATING_READER(DeprecatedReader):
    _parseMethod = (valParse,)
