########################################################################
#
# File Name:            Util.py
#
# Documentation:        http://docs.4suite.org/4XSLT/Util.py.html
#
"""
General Utilities for XPath apps.
WWW: http://4suite.org/4XSLT        e-mail: support@4suite.org

Copyright (c) 2000-2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import os, glob, string
from Ft.Xml import SplitQName
from Ft.Xml import Domlette, XML_NAMESPACE, XMLNS_NAMESPACE, EMPTY_NAMESPACE
from xml.dom import Node
from Ft.Xml.XPath import g_xpathRecognizedNodes
from Ft.Xml.XPath.NamespaceNode import NamespaceNode
from Ft.Xml.XPath import RuntimeException

g_documentOrderIndex = {}
g_documentIdIndex = {}

def GetElementById(document, name):
    if not g_documentIdIndex.has_key(id(document)):
        # First time through for this document, build ID index
        g_documentIdIndex[id(document)] = mapping = {}
        ElementsById(document.documentElement, mapping)

    elements = g_documentIdIndex[id(document)].get(name, [None])
    if len(elements) > 1:
        raise ValueError("ID not unique")
    return elements[0]


def ElementsById(element, mapping):
    idattr = element.getAttributeNodeNS(EMPTY_NAMESPACE, 'ID') or \
             element.getAttributeNodeNS(EMPTY_NAMESPACE, 'id')
    if idattr:
        mapping.setdefault(idattr.value, []).append(idattr.ownerElement)

    for element in filter(lambda node: node.nodeType == Node.ELEMENT_NODE,
                          element.childNodes):
        ElementsById(element, mapping)
    return


def IndexDocument(doc):
    global g_documentOrderIndex
    if g_documentOrderIndex.has_key(id(doc)):
        return
    mapping = {}
    __IndexNode(doc, 0, mapping)
    g_documentOrderIndex[id(doc)] = mapping


def FreeDocumentIndex(doc):
    global g_documentOrderIndex
    try:
        del g_documentOrderIndex[id(doc)]
    except KeyError:
        pass
    try:
        del g_documentIdIndex[id(doc)]
    except KeyError:
        pass
    return

def SortDocOrder(context, nodeset):
    if len(nodeset) <= 1:
        return nodeset

    # Determine if the nodeset contains mixed documents
    documents = {}
    for node in nodeset:
        documents[node.rootNode] = 1

    if len(documents) > 1:
        # Mixed documents, split the nodesets based on document
        nodesets = []
        documents = documents.keys()
        documents.sort(context.compareDocuments)
        for document in documents:
            nodesets.append(filter(lambda node, doc=document:
                                   node.rootNode == doc,
                                   nodeset))

        sorted = []
        for nodeset in nodesets:
            decorated = [ (node.docIndex, node) for node in nodeset ]
            decorated.sort()
            sorted.extend([ item[1] for item in decorated ])

    else:
        decorated = [ (node.docIndex, node) for node in nodeset ]
        decorated.sort()
        sorted = [ item[1] for item in decorated ]

    return sorted


def ExpandQName(qname, refNode=None, namespaces=None):
    '''
    Expand the given QName in the context of the given node,
    or in the given namespace dictionary
    '''
    nss = {}
    if refNode:
        nss = Domlette.GetAllNs(refNode)
    elif namespaces:
        nss = namespaces
    (prefix, local) = SplitQName(qname)
    #We're not to use the default namespace
    if prefix:
        try:
            split_name = (nss[prefix], local)
        except KeyError:
            raise RuntimeException(RuntimeException.UNDEFINED_PREFIX,
                                   prefix)
    else:
        split_name = (EMPTY_NAMESPACE, local)
    return split_name


def __IndexNode(node, curIndex, mapping):
    if node.nodeType in g_xpathRecognizedNodes:
        #Add this node
        mapping[id(node)] = curIndex
        curIndex = curIndex + 1

        if node.nodeType == Node.ELEMENT_NODE:
            # Leave a space for namespace nodes
            curIndex += 1

            # It is OK for attributes to be unordered
            for attr in node.attributes.values():
                mapping[id(attr)] = curIndex
                curIndex += 1

        for childNode in node.childNodes:
            curIndex = __IndexNode(childNode, curIndex, mapping)

    return curIndex


def GetIndex(node):
    nid = id(node)
    docId = id(node.rootNode)
    try:
        ix = g_documentOrderIndex[docId][nid]
    except KeyError:
        #Must be namespace node in there
        if isinstance(node, NamespaceNode):
            nid = id(node.parentNode)
            ix = g_documentOrderIndex[docId][nid] + 1
    return ix


def __recurseSort(test, toSort):
    """Check whether any of the nodes in toSort are in the list test, and if so, sort them into the result list"""
    result = []
    for node in test:
        toSort = filter(lambda x, n=node: x != n, toSort)
        if node in toSort:
            result.append(node)
        #See if node has attributes
        if node.nodeType == Node.ELEMENT_NODE:
            attrList = node.attributes.values()
            #FIXME: Optimize by unrolling this level of recursion
            result = result + __recurseSort(attrList, toSort)
            if not toSort:
                #Exit early
                break
        #See if any of t's children are in toSort
        result = result + __recurseSort(node.childNodes, toSort)
        if not toSort:
            #Exit early
            break
    return result


def NormalizeNode(node):
    """NormalizeNode is used to prepare a DOM for XPath evaluation.

    1.  Convert CDATA Sections to Text Nodes.
    2.  Normalize all text nodes
    """
    node = node.firstChild
    while node:
        if node.nodeType == Node.CDATA_SECTION_NODE:
            # If followed by a text node, add this data to it
            if node.nextSibling and node.nextSibling.nodeType == Node.TEXT_NODE:
                node.nextSibling.insertData(0, node.data)
            elif node.data:
                # Replace this node with a new text node
                text = node.ownerDocument.createTextNode(node.data)
                node.parentNode.replaceChild(text, node)
                node = text
            else:
                # It is empty, get rid of it
                next = node.nextSibling
                node.parentNode.removeChild(node)
                node = next
                # Just in case it is None
                continue
        elif node.nodeType == Node.TEXT_NODE:
            next = node.nextSibling
            while next and next.nodeType in [Node.TEXT_NODE,
                                             Node.CDATA_SECTION_NODE]:
                node.appendData(next.data)
                node.parentNode.removeChild(next)
                next = node.nextSibling
            if not node.data:
                # Remove any empty text nodes
                next = node.nextSibling
                node.parentNode.removeChild(node)
                node = next
                # Just in case it is None
                continue
        elif node.nodeType == Node.ELEMENT_NODE:
            for attr in node.attributes.values():
                if len(attr.childNodes) > 1:
                    NormalizeNode(attr)
            NormalizeNode(node)
        node = node.nextSibling
    return

