#!/usr/bin/env python
########################################################################
#
# File Name:            _4xpath.py
#
# Documentation:        http://docs.4suite.org/4XPath/4xpath.py.html
#
"""
Command-line invocation of the 4XPath
WWW: http://4suite.org/4XPath        e-mail: support@4suite.org

Copyright 1999-2002 Fourthought, Inc., USA.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import re, os, sys, traceback, types
from Ft.Xml import XPath, InputSource
from Ft.Xml import __version__
from Ft.Xml.XPath.Util import ExpandQName
from Ft.Xml.XPath import Conversions
from Ft.Lib import boolean, Uri

__doc__ = """4XPath command-line application"""

g_paramBindingPattern = re.compile(r"([\d\D_\.\-]*:?[\d\D_\.\-]+)=(.*)")
g_nssBindingPattern = re.compile(r"([\d\D_\.\-]*:?[\d\D_\.\-]+)=(.*)")

from Ft import MAX_PYTHON_RECURSION_DEPTH
sys.setrecursionlimit(MAX_PYTHON_RECURSION_DEPTH)

def Run(options,args):

    if options.has_key('version'):
        print '4XPath, from 4Suite %s' % __version__
        return

    if not args.has_key('source-uri'):
        print 'A source URI argument is required. See "4xpath -h" for usage info.'
        return

    defs = options.get('define', [])
    if type(defs) != type([]):
        defs = [defs]

    namespaces = options.get('namespace', [])
    if type(namespaces) != type([]):
        namespaces = [namespaces]

    processorNss = {}
    for n in namespaces:
        match = g_nssBindingPattern.match(n)
        processorNss[match.group(1)] = match.group(2)

    top_level_params = {}
    for d in defs:
        match = g_paramBindingPattern.match(d)
        name = ExpandQName(match.group(1),namespaces = processorNss)
        top_level_params[name] = match.group(2)



    stringValue = options.has_key('string')
    stacktrace_on_error = options.has_key('stacktrace-on-error')

    from Ft.Xml import Domlette
    reader = Domlette.NonvalidatingReader

    try:

        compExpr = XPath.Compile(args['expression'])

        source = args['source-uri']
        if source == '-':
            isrc = InputSource.InputSource(sys.stdin,'.')
        else:
            source = Uri.OsPathToUri(source, attemptAbsolute=1)
            isrc = InputSource.DefaultFactory.fromUri(source)
            
        dom = reader.parse(isrc)

        context = XPath.Context.Context(dom,
                                        processorNss = processorNss,
                                        varBindings = top_level_params,
                                        extModuleList = os.environ.get("EXTMODULES","").split(":"))
                                        
        res = compExpr.evaluate(context)
        if stringValue:
            res = Conversions.StringValue(res)
        if type(res) in [types.StringType,types.UnicodeType]:
            print "String Results: "
            print res
        elif type(res) in [types.IntType,types.FloatType]:
            print "Number Results: "
            print res
        elif type(res) == boolean.BooleanType:
            print "Boolean Results: "
            print res
        elif type(res) == types.ListType:
            print "Node Set: "
            for node in res:
                print node
        else:
            print "Unknown Type: "
            print res
    except (XPath.RuntimeException, XPath.CompiletimeException), error:
        if stacktrace_on_error:
            traceback.print_exc(1000, sys.stderr)
        raise SystemExit(str(error))
    except Exception, error:
        if stacktrace_on_error:
            traceback.print_exc(1000, sys.stderr)
        raise SystemExit(str(error))


from Ft.Lib.CommandLine import Options, CommandLineApp, Arguments, Command


class XPathCommandLineApp(CommandLineApp.CommandLineApp):
    def __init__(self):
        CommandLineApp.CommandLineApp.__init__(
            self,
            '4xpath',
            '4XPath version %s' % __version__,
            __doc__,
            [],
            ourOptions = Options.Options([Options.Option('V',
                                                         'version',
                                                         'Display program version and exit',
                                                         ),
                                          Options.Option('D',
                                                         'define=NAME=VALUE',
                                                         'Bind a top-level parameter'
                                                         ),
                                          Options.Option('N',
                                                         'namespace=PREFIX=NAMESPACE',
                                                         'Define a namespace to beused for prefix resolution'
                                                         ),
                                          Options.Option('e',
                                                         'stacktrace-on-error',
                                                         'Display a stack trace when an error occurs',
                                                         ),
                                          Options.Option(None,
                                                         'string',
                                                         'Print out the string value of the results',
                                                         ),
                                          #Options.Option('d',
                                          #               'debug',
                                          #               'Execute 4xslt in the command line debugger',
                                          #               ),
                                          ]),
            enableShowCommands = 0
            )

        self.function = Run
        self.arguments = [Arguments.OptionalArgument(
            'source-uri',
            'The URI of the source XML document. Required unless running "4xpath -V".'\
            ' Note: if you use "-" as the name of the source document, the source will'\
            ' instead be read from standard input.',
            str),
                          Arguments.RequiredArgument('expression',
                                                      'The XPath expression to evaluate',
                                                      str),
                          ]


    def validate_arguments(self,args):
        return Command.Command.validate_arguments(self,args)



def Register():
    return XPathCommandLineApp()
