#!/usr/bin/env python
########################################################################
#
# File Name:            _4xml.py
#
# Documentation:        http://docs.4suite.org/4xml.py.html
#
"""
Command-line invocation of the 4XML processor
WWW: http://4suite.org        e-mail: support@4suite.org

Copyright 1999-2002 Fourthought, Inc., USA.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

##Usage:
##  4xml [--help] [-i | --in-place] [-o | --out-file=<out-file>]
##       [ --noxinclude ] source-uri

##Options:
##  -h, --help               show detailed help message
##  -i, --in-place           Update the source file in place
##  -o, --out-file=<out-file>  Write the result to the given output file.
##  --noxinclude             Suppress XInclude processing

##Arguments:
##  source-uri   The URI of the source XML document for parsing.
##               If you use "-" as the name of the source document,
##               the source will instead be read from standard input.
##               "-" cannot be specified as the source URI
##               if the "-i" option is used.

import re, os, sys, traceback, codecs
#import string, getopt, cStringIO
from Ft.Xml import __version__
from Ft.Xml import InputSource
from Ft.Lib import Uri

__doc__ = """4XML command-line application"""


def Run(options, args):
    if options.has_key('version'):
        print '4XML, from 4Suite %s' % __version__
        return

    if not args.has_key('source-uri'):
        print 'A source URI argument is required. See "4xml -h" for usage info.'
        return

    xinclude = not options.has_key('noxinclude')
    output = options.has_key('outfile') and open(options['outfile'], 'w') or sys.stdout
    validate_flag = options.has_key('validate')
    pretty = options.has_key('pretty')
    as_html = options.has_key('html')
    encoding = options.get('encoding', 'utf-8')
    rng_file = options.get('rng')

    # Do the file checking here; the rest of 4XML is uri based
    source = args['source-uri']
    if os.path.isfile(source):
        #source = urllib.pathname2url(source)
        source = Uri.OsPathToUri(source, attemptAbsolute=1)

    if rng_file and os.path.isfile(rng_file):
        rng_file = Uri.OsPathToUri(rng_file, attemptAbsolute=1)

    from Ft.Xml import Domlette

    if validate_flag:
        reader = Domlette.ValidatingReader
    else:
        reader = Domlette.NonvalidatingReader

    try:
        if source == '-':
            isrc = InputSource.InputSource(
                sys.stdin, '.', processIncludes=xinclude)
        else:
            isrc = InputSource.DefaultFactory.fromUri(
                source, processIncludes=xinclude)
        doc = reader.parse(isrc)
        if rng_file:
            from Ft.Xml.Xvif import RelaxNgValidator, RngInvalid
            rng_isrc = InputSource.DefaultFactory.fromUri(
                rng_file, processIncludes=xinclude)
            validator = RelaxNgValidator(rng_isrc)
            result = validator.validateNode(doc)
            if not result.nullable():
                raise RngInvalid(result)
        from Ft.Xml.Domlette import Print, PrettyPrint
        if pretty:
            PrettyPrint(doc, output, encoding, as_html)
        else:
            Print(doc, output, encoding, as_html)
    except Exception, e:
        import traceback; traceback.print_exc(1000, sys.stderr)
        raise

    output.close()


from Ft.Lib.CommandLine import Options, CommandLineApp, Arguments, Command


class XmlCommandLineApp(CommandLineApp.CommandLineApp):
    def __init__(self):
        CommandLineApp.CommandLineApp.__init__(
            self,
            '4xml',
            '4XML version %s' % __version__,
            __doc__,
            [],
            ourOptions = Options.Options([Options.Option('V',
                                                         'version',
                                                         'Display program version and exit',
                                                         ),
                                          Options.Option('v',
                                                         'validate',
                                                         'DTD validate the input file as it is being parsed',
                                                         ),
                                          Options.Option('e',
                                                         'encoding=ENC',
                                                         'The encoding to use for output',
                                                         ),
                                          Options.Option('o',
                                                         'outfile=FILE',
                                                         'Direct transformation output to FILE '\
                                                         '(file will be overwritten if it exists)'
                                                         ),
                                          Options.Option('p',
                                                         'pretty',
                                                         'Pretty-print the result',
                                                         ),
                                          Options.Option(None,
                                                         'html',
                                                         'Use HTML mode when pretty-printing (emit XHTML as HTML)',
                                                         ),
                                          Options.Option(None,
                                                         'noxinclude',
                                                         'Do not expand XIncludes in source or stylesheet'
                                                         ),
                                          Options.Option(None,
                                                         'rng=FILE',
                                                         'Apply RELAX NG from the given file (technically RNG with XVIF features)'
                                                         ),
                                          ]),
            enableShowCommands = 0
            )

        self.function = Run
        self.arguments = [Arguments.OptionalArgument(
            'source-uri',
            'The URI of the XML document to parse.  Required unless running "4xml -V".'\
            ' Note: if you use "-" as the name of the source document, the source will'\
            ' instead be read from standard input.',
            str),
                          ]


    def validate_arguments(self,args):
        return Command.Command.validate_arguments(self,args)


def Register():
    return XmlCommandLineApp()

